from __future__ import annotations

import gzip
import html
import os
from dataclasses import dataclass
from functools import lru_cache
from typing import List, Union, Iterable

import cv2
import ftfy
import numpy as np
import regex as re
from PIL import Image
from onnxruntime import InferenceSession


@lru_cache()
def default_bpe():
    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "data", "bpe_simple_vocab_16e6.txt.gz"
    )


@lru_cache()
def bytes_to_unicode():
    """
    Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings.
    This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
    This is a signficant percentage of your normal, say, 32K bpe vocab.
    To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
    And avoids mapping to whitespace/control characters the bpe code barfs on.
    """
    bs = (
        list(range(ord("!"), ord("~") + 1))
        + list(range(ord("¡"), ord("¬") + 1))
        + list(range(ord("®"), ord("ÿ") + 1))
    )
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))


def get_pairs(word):
    """Return set of symbol pairs in a word.
    Word is represented as tuple of symbols (symbols being variable-length strings).
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs


def basic_clean(text):
    text = ftfy.fix_text(text)
    text = html.unescape(html.unescape(text))
    return text.strip()


def whitespace_clean(text):
    text = re.sub(r"\s+", " ", text)
    text = text.strip()
    return text


class Tokenizer:
    def __init__(self, bpe_path: str = default_bpe()):
        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v: k for k, v in self.byte_encoder.items()}
        merges = gzip.open(bpe_path).read().decode("utf-8").split("\n")
        merges = merges[1 : 49152 - 256 - 2 + 1]
        merges = [tuple(merge.split()) for merge in merges]
        vocab = list(bytes_to_unicode().values())
        vocab = vocab + [v + "</w>" for v in vocab]
        for merge in merges:
            vocab.append("".join(merge))
        vocab.extend(["<|startoftext|>", "<|endoftext|>"])
        self.encoder = dict(zip(vocab, range(len(vocab))))
        self.decoder = {v: k for k, v in self.encoder.items()}
        self.bpe_ranks = dict(zip(merges, range(len(merges))))
        self.cache = {"<|startoftext|>": "<|startoftext|>", "<|endoftext|>": "<|endoftext|>"}
        self.pat = re.compile(
            r"""<\|startoftext\|>|<\|endoftext\|>|'s|'t|'re|'ve|'m|'ll|'d|[\p{L}]+|[\p{N}]|[^\s\p{L}\p{N}]+""",
            re.IGNORECASE,
        )

    def bpe(self, token: str) -> str:
        if token in self.cache:
            return self.cache[token]
        word = tuple(token[:-1]) + (token[-1] + "</w>",)
        pairs = get_pairs(word)

        if not pairs:
            return token + "</w>"

        while True:
            bigram = min(pairs, key=lambda pair: self.bpe_ranks.get(pair, float("inf")))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)
        word = " ".join(word)
        self.cache[token] = word
        return word

    def encode(self, text: str) -> List[int]:
        bpe_tokens = []
        text = whitespace_clean(basic_clean(text)).lower()
        for token in re.findall(self.pat, text):
            token = "".join(self.byte_encoder[b] for b in token.encode("utf-8"))
            bpe_tokens.extend(self.encoder[bpe_token] for bpe_token in self.bpe(token).split(" "))
        return bpe_tokens

    def decode(self, tokens: List[int]) -> str:
        text = "".join([self.decoder[token] for token in tokens])
        text = (
            bytearray([self.byte_decoder[c] for c in text])
            .decode("utf-8", errors="replace")
            .replace("</w>", " ")
        )
        return text

    def __call__(
        self, texts: Union[str, Iterable[str]], context_length: int = 77, *args, **kwargs
    ) -> np.array:
        if isinstance(texts, str):
            texts = [texts]

        sot_token = self.encoder["<|startoftext|>"]
        eot_token = self.encoder["<|endoftext|>"]
        all_tokens = [[sot_token] + self.encode(text) + [eot_token] for text in texts]
        result = np.zeros((len(all_tokens), context_length), dtype=np.int32)

        for i, tokens in enumerate(all_tokens):
            if len(tokens) > context_length:
                tokens = tokens[:context_length]
                tokens[-1] = eot_token
            result[i, : len(tokens)] = np.array(tokens)

        return result


class Preprocessor:
    """
    Our approach to the CLIP `preprocess` neural net that does not rely on PyTorch.
    The two approaches fully match.
    """

    # Fixed variables that ensure the correct output shapes and values for the `Model` class.
    CLIP_INPUT_SIZE = 224
    # Normalization constants taken from original CLIP:
    # https://github.com/openai/CLIP/blob/3702849800aa56e2223035bccd1c6ef91c704ca8/clip/clip.py#L85
    NORM_MEAN = np.array([0.48145466, 0.4578275, 0.40821073]).reshape((1, 1, 3))
    NORM_STD = np.array([0.26862954, 0.26130258, 0.27577711]).reshape((1, 1, 3))

    @staticmethod
    def _crop_and_resize(img: np.ndarray) -> np.ndarray:
        """Resize and crop an image to a square, preserving the aspect ratio."""

        # Current height and width
        h, w = img.shape[0:2]

        if h * w == 0:
            raise ValueError(
                f"Height and width of the image should both be non-zero but got shape {h, w}"
            )

        target_size = Preprocessor.CLIP_INPUT_SIZE

        # Resize so that the smaller dimension matches the required input size.
        # Matches PyTorch:
        # https://github.com/pytorch/vision/blob/7cf0f4cc1801ff1892007c7a11f7c35d8dfb7fd0/torchvision/transforms/functional.py#L366
        if h < w:
            resized_h = target_size
            resized_w = int(resized_h * w / h)
        else:
            resized_w = target_size
            resized_h = int(resized_w * h / w)

        # PIL resizing behaves slightly differently than OpenCV because of
        # antialiasing. See also
        # https://pytorch.org/vision/main/generated/torchvision.transforms.functional.resize.html
        # CLIP uses PIL, so we do too to match its results. But if you don't
        # want to have PIL as a dependency, feel free to change the code to
        # use the other branch.
        use_pil_for_resizing = True

        if use_pil_for_resizing:
            # https://github.com/pytorch/vision/blob/7cf0f4cc1801ff1892007c7a11f7c35d8dfb7fd0/torchvision/transforms/functional_pil.py#L240
            # We're working with float images but PIL uses uint8, so convert
            # there and back again afterwards
            img_pil = Image.fromarray((img * 255).astype(np.uint8))
            img_pil = img_pil.resize((resized_w, resized_h), resample=Image.BICUBIC)
            img = np.array(img_pil).astype(np.float32) / 255
        else:
            img = cv2.resize(img, (resized_w, resized_h), interpolation=cv2.INTER_CUBIC)

        # Now crop to a square
        y_from = (resized_h - target_size) // 2
        x_from = (resized_w - target_size) // 2
        img = img[y_from : y_from + target_size, x_from : x_from + target_size, :]

        return img

    @staticmethod
    def _image_to_float_array(img: Union[Image.Image, np.ndarray]):
        """Converts a PIL image or a NumPy array to standard form.

        Standard form means:
        - the shape is (H, W, 3)
        - the dtype is np.float32
        - all values are in [0, 1]
        - there are no NaN values

        Args:
            img: The image to convert.

        Returns:
            The image converted to a NumPy array in standard form.

        Raises:
            ValueError if the image is invalid (wrong shape, invalid
                values...).
        """
        if not isinstance(img, (Image.Image, np.ndarray)):
            raise TypeError(f"Expected PIL Image or np.ndarray but instead got {type(img)}")

        if isinstance(img, Image.Image):
            # Convert to NumPy
            img = np.array(img)

        if len(img.shape) > 3:
            raise ValueError(
                f"The image should have 2 or 3 dimensions but got {len(img.shape)} dimensions"
            )
        if len(img.shape) == 3 and img.shape[2] != 3:
            raise ValueError(
                f"Expected 3-channel RGB image but got image with {img.shape[2]} channels"
            )

        # Handle grayscale
        if len(img.shape) == 2:
            # The model doesn't support HxWx1 images as input
            img = np.expand_dims(img, axis=2)  # HxWx1
            img = np.concatenate((img,) * 3, axis=2)  # HxWx3

        # At this point, `img` has the shape (H, W, 3).

        if np.min(img) < 0:
            raise ValueError(
                "Images should have non-negative pixel values, "
                f"but the minimum value is {np.min(img)}"
            )

        if np.issubdtype(img.dtype, np.floating):
            if np.max(img) > 1:
                raise ValueError(
                    "Images with a floating dtype should have values "
                    f"in [0, 1], but the maximum value is {np.max(img)}"
                )
            img = img.astype(np.float32)
        elif np.issubdtype(img.dtype, np.integer):
            if np.max(img) > 255:
                raise ValueError(
                    "Images with an integer dtype should have values "
                    f"in [0, 255], but the maximum value is {np.max(img)}"
                )
            img = img.astype(np.float32) / 255
            img = np.clip(img, 0, 1)  # In case of rounding errors
        else:
            raise ValueError(f"The image has an unsupported dtype: {img.dtype}.")

        if np.isnan(img).any():
            raise ValueError(f"The image contains NaN values.")

        try:
            # These should never trigger, but let's do a sanity check
            assert np.min(img) >= 0
            assert np.max(img) <= 1
            assert img.dtype == np.float32
            assert len(img.shape) == 3
            assert img.shape[2] == 3
        except AssertionError as e:
            raise RuntimeError(
                "Internal preprocessing error. The image does not have the expected format."
            ) from e

        return img

    def __call__(self, img: Union[Image.Image, np.ndarray], *args, **kwargs) -> np.ndarray:
        """Preprocesses the images like CLIP's preprocess() function:

        Args:
            img: PIL image or numpy array

        Returns:
            img: numpy image after resizing, center cropping and normalization.
        """
        img = Preprocessor._image_to_float_array(img)

        img = Preprocessor._crop_and_resize(img)

        # Normalize channels
        img = (img - Preprocessor.NORM_MEAN) / Preprocessor.NORM_STD

        # Mimic the pytorch tensor format for Model class
        img = np.transpose(img, (2, 0, 1))
        img = np.expand_dims(img, 0)
        img = img.astype(np.float32)

        return img


@dataclass
class MossCLIP:
    visual_session: InferenceSession
    textual_session: InferenceSession

    _tokenizer = None
    _preprocessor = None

    def __post_init__(self):
        self._tokenizer = Tokenizer()
        self._preprocessor = Preprocessor()

    @classmethod
    def from_pluggable_model(cls, visual_model: InferenceSession, textual_model: InferenceSession):
        return cls(visual_session=visual_model, textual_session=textual_model)

    def encode_image(self, images: Iterable[Image.Image | np.ndarray]) -> np.ndarray:
        """
        Compute the embeddings for a list of images.

        :param images:
            A list of images to run on. Each image must be a 3-channel(RGB) image.
            Can be any size, as the preprocessing step will resize each image to size (224, 224).
        :return:
            An array of embeddings of shape (len(images), embedding_size).

        """
        images = [self._preprocessor(image) for image in images]
        batch = np.concatenate(images)
        input_name = self.visual_session.get_inputs()[0].name
        return self.visual_session.run(None, {input_name: batch})[0]

    def encode_text(self, texts: Iterable[str]) -> np.ndarray:
        """
        Compute the embeddings for a list of texts.

        :param texts:
            A list of texts to run on. Each entry can be at most 77 characters.
        :return:
            An array of embeddings of shape (len(texts), embedding_size).
        """
        text = self._tokenizer(texts)
        input_name = self.textual_session.get_inputs()[0].name
        return self.textual_session.run(None, {input_name: text})[0]

    def __call__(
        self, images: Iterable[Image.Image | np.ndarray], candidate_labels, *args, **kwargs
    ):
        """

        :param images:
        :param candidate_labels:
        :param args:
        :param kwargs:
        :return:
            A list of dictionaries containing result, one dictionary per proposed label. The dictionaries contain the
            following keys:

            - **label** (`str`) -- The label identified by the model. It is one of the suggested `candidate_label`.
            - **score** (`float`) -- The score attributed by the model for that label (between 0 and 1).
        """

        def softmax(x):
            return np.exp(x) / np.sum(np.exp(x), axis=1, keepdims=True)

        image_features = self.encode_image(images)
        text_features = self.encode_text(candidate_labels)

        image_features /= np.linalg.norm(image_features, axis=1, keepdims=True)
        text_features /= np.linalg.norm(text_features, axis=1, keepdims=True)

        text_probs = 100 * image_features @ text_features.T
        text_probs = softmax(text_probs)

        result = [
            {"score": score, "label": label}
            for score, label in sorted(zip(text_probs[0], candidate_labels), key=lambda x: -x[0])
        ]
        return result



# -*- coding: utf-8 -*-
# Time       : 2023/8/19 17:53
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

from contextlib import suppress
from pathlib import Path
from typing import List, Dict, Tuple

import cv2
from PIL import Image
from loguru import logger

from hcaptcha_challenger.components.prompt_handler import handle
from hcaptcha_challenger.components.zero_shot_image_classifier import (
    ZeroShotImageClassifier,
    register_pipline,
)
from hcaptcha_challenger.onnx.modelhub import ModelHub, DataLake
from hcaptcha_challenger.onnx.resnet import ResNetControl


def rank_models(
    nested_models: List[str], example_paths: List[Path], modelhub: ModelHub
) -> Tuple[ResNetControl, str] | None:
    # {{< Rank ResNet Models >}}
    rank_ladder = []

    for example_path in example_paths:
        img_stream = example_path.read_bytes()
        for model_name in reversed(nested_models):
            if (net := modelhub.match_net(focus_name=model_name)) is None:
                return
            control = ResNetControl.from_pluggable_model(net)
            result_, proba = control.execute(img_stream, proba=True)
            if result_:
                rank_ladder.append([control, model_name, proba])
                if proba[0] > 0.87:
                    break

    # {{< Catch-all Rules >}}
    if rank_ladder:
        alts = sorted(rank_ladder, key=lambda x: x[-1][0], reverse=True)
        best_model, model_name = alts[0][0], alts[0][1]
        return best_model, model_name


class Classifier:
    def __init__(
        self,
        *,
        modelhub: ModelHub | None = None,
        clip_model=None,
        datalake_post: Dict[str, Dict[str, List[str]]] | None = None,
        **kwargs,
    ):
        self.modelhub = modelhub or ModelHub.from_github_repo()
        if not self.modelhub.label_alias:
            self.modelhub.parse_objects()

        if isinstance(datalake_post, dict):
            for prompt, serialized_binary in datalake_post.items():
                self.modelhub.datalake[prompt] = DataLake.from_serialized(serialized_binary)

        self.clip_model = clip_model

        self.response: List[bool | None] = []
        self.prompt: str = ""
        self.label: str = ""
        self.model_name: str = ""

    def _parse_label(self, prompt: str):
        self.prompt = prompt
        self.label = handle(self.prompt)

    def rank_models(
        self, nested_models: List[str], example_paths: List[Path]
    ) -> ResNetControl | None:
        result = rank_models(nested_models, example_paths, self.modelhub)
        if result and isinstance(result, tuple):
            best_model, model_name = result
            self.model_name = model_name
            return best_model

    def inference(self, images: List[Path | bytes], control: ResNetControl):
        for image in images:
            try:
                if isinstance(image, Path):
                    if not image.exists():
                        self.response.append(None)
                        continue
                    image = image.read_bytes()
                if isinstance(image, bytes):
                    result = control.execute(image)
                    self.response.append(result)
                else:
                    self.response.append(None)
            except Exception as err:
                logger.debug(str(err), prompt=self.prompt)
                self.response.append(None)

    def inference_by_clip(self, image_paths: List[Path]):
        dl = self.modelhub.datalake.get(self.label)
        if not dl:
            dl = DataLake.from_challenge_prompt(raw_prompt=self.label)
        tool = ZeroShotImageClassifier.from_datalake(dl)

        # Default to `RESNET.OPENAI` perf_counter 1.794s
        model = self.clip_model or register_pipline(self.modelhub)
        self.model_name = self.modelhub.DEFAULT_CLIP_VISUAL_MODEL

        for image_path in image_paths:
            try:
                if not isinstance(image_path, Path):
                    raise TypeError(
                        "Please pass in the pathlib.Path object, "
                        "you don't need to set it specifically for bytes in advance. "
                        f"- type={type(image_path)}"
                    )
                if not image_path.exists():
                    raise FileNotFoundError(f"ChallengeImage not found - path={image_path}")
                # self-supervised image classification
                results = tool(model, image=Image.open(image_path))
                trusted = results[0]["label"] in tool.positive_labels
                self.response.append(trusted)
            except Exception as err:
                logger.debug(str(err), prompt=self.prompt)
                self.response.append(None)

        # Pop the temporarily inserted model and free up memory
        if not self.clip_model:
            self.modelhub.unplug()

    def execute(
        self,
        prompt: str,
        image_paths: List[Path],
        example_paths: List[Path] = None,
        *,
        self_supervised: bool | None = True,
    ) -> List[bool | None]:
        self.response = []

        self._parse_label(prompt)

        # Match: model ranker
        if nested_models := self.modelhub.nested_categories.get(self.label, []):
            if model := self.rank_models(nested_models, example_paths):
                self.inference(image_paths, model)
        # Match: common binary classification
        elif focus_label := self.modelhub.label_alias.get(self.label):
            if focus_label.endswith(".onnx"):
                self.model_name = focus_label
            else:
                self.model_name = f"{focus_label}.onnx"
            net = self.modelhub.match_net(self.model_name)
            control = ResNetControl.from_pluggable_model(net)
            self.inference(image_paths, control)
        elif self_supervised:
            self.inference_by_clip(image_paths)
        # Match: Unknown cases
        else:
            logger.debug("Types of challenges not yet scheduled", label=self.label, prompt=prompt)

        return self.response


class LocalBinaryClassifier:
    def __init__(self, model_path: Path):
        if not isinstance(model_path, Path) or not model_path.exists():
            raise FileNotFoundError
        self.model_path = model_path
        net = cv2.dnn.readNetFromONNX(str(model_path))
        self.model = ResNetControl.from_pluggable_model(net)

    def parse_once(self, image: bytes) -> bool | None:
        with suppress(Exception):
            return self.model.execute(image)



# -*- coding: utf-8 -*-
# Time       : 2022/2/15 17:43
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Iterable
from urllib.parse import urlparse

from hcaptcha_challenger.components.image_classifier import Classifier as BinaryClassifier
from hcaptcha_challenger.components.image_classifier import LocalBinaryClassifier
from hcaptcha_challenger.components.image_label_area_select import AreaSelector
from hcaptcha_challenger.components.prompt_handler import (
    label_cleaning,
    diagnose_task,
    split_prompt_message,
    prompt2task,
    handle,
)
from hcaptcha_challenger.components.zero_shot_image_classifier import (
    ZeroShotImageClassifier,
    DataLake,
    register_pipline,
)
from hcaptcha_challenger.onnx.modelhub import ModelHub
from hcaptcha_challenger.onnx.resnet import ResNetControl
from hcaptcha_challenger.onnx.yolo import YOLOv8
from hcaptcha_challenger.onnx.yolo import YOLOv8Seg
from hcaptcha_challenger.utils import init_log

__all__ = [
    "BinaryClassifier",
    "LocalBinaryClassifier",
    "ZeroShotImageClassifier",
    "register_pipline",
    "DataLake",
    "AreaSelector",
    "label_cleaning",
    "diagnose_task",
    "split_prompt_message",
    "prompt2task",
    "handle",
    "ModelHub",
    "ResNetControl",
    "YOLOv8",
    "YOLOv8Seg",
    "install",
]


@dataclass
class Project:
    at_dir = Path(__file__).parent
    logs = at_dir.joinpath("logs")


project = Project()

init_log(
    runtime=project.logs.joinpath("runtime.log"),
    error=project.logs.joinpath("error.log"),
    serialize=project.logs.joinpath("serialize.log"),
)


def install(
    upgrade: bool | None = False,
    username: str = "QIN2DIM",
    lang: str = "en",
    flush_yolo: bool | Iterable[str] = False,
    pypi: bool = False,
    clip: bool = False,
    **kwargs,
):
    if pypi is True:
        from hcaptcha_challenger.utils import PyPI

        PyPI("hcaptcha-challenger").install()

    modelhub = ModelHub.from_github_repo(username=username, lang=lang)
    modelhub.pull_objects(upgrade=upgrade)
    modelhub.assets.flush_runtime_assets(upgrade=upgrade)

    if clip is True:
        from hcaptcha_challenger.components.zero_shot_image_classifier import register_pipline

        register_pipline(modelhub, install_only=True)

    if flush_yolo is not None:
        modelhub.parse_objects()

        if isinstance(flush_yolo, bool) and flush_yolo:
            flush_yolo = [modelhub.circle_segment_model]
        if isinstance(flush_yolo, Iterable):
            pending_models = []
            for model_name in flush_yolo:
                if model_name in modelhub.ashes_of_war:
                    modelhub.match_net(model_name, install_only=True)
                    pending_models.append(model_name)
            return pending_models


def set_reverse_proxy(https_cdn: str):
    parser = urlparse(https_cdn)
    if parser.netloc and parser.scheme.startswith("https"):
        ModelHub.CDN_PREFIX = https_cdn



# -*- coding: utf-8 -*-
# Time       : 2023/9/29 14:19
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from typing import Tuple, List, Literal

import cv2
import numpy as np
from skimage.metrics import structural_similarity as compare_ssim


def limited_radius(img) -> int:
    # x1, y1, x2, y2 = 90, 200, 413, 523
    r = 48 if 512 in img.shape else 32
    return r


def annotate_objects(image_path: str):
    img = cv2.imread(image_path)
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    circles = cv2.HoughCircles(
        gray, cv2.HOUGH_GRADIENT, 1, 66, param1=100, param2=40, minRadius=5, maxRadius=65
    )

    if np.all(circles):
        circles = np.uint16(np.around(circles))
        ix = []
        for i in circles[0, :]:
            lr = limited_radius(img)
            if i[2] > lr + 2:
                continue
            i[2] = lr
            ix.append(i)
        return img, ix

    return img, None


def _build_mask(img: np.ndarray, circles: List[List[int]], lookup: Literal["color", "object"]):
    mask_images = []

    for circle in circles:
        x, y, r = circle
        mask = np.zeros(img.shape[:2], dtype=np.uint8)
        mask = cv2.circle(mask, (x, y), r, (255, 255, 0), -1)
        mask_img = cv2.bitwise_and(img, img, mask=mask)
        mask_img = mask_img[y - r : y + r, x - r : x + r]
        if lookup == "color":
            mask_img[np.where((mask_img == [0, 0, 0]).all(axis=2))] = [255, 255, 255]
        mask_images.append(mask_img)

    # uniform size, brightness, contrast, etc.
    max_size = max([mask_img.shape[0] for mask_img in mask_images])
    mask_images = [cv2.resize(mask_img, (max_size, max_size)) for mask_img in mask_images]

    return mask_images


def find_unique_object(img: np.ndarray, circles: List[List[int]]) -> Tuple[int, int, int]:
    mask_images = _build_mask(img, circles, lookup="object")

    similarity = []

    for i, mask_img in enumerate(mask_images):
        sig_sim = []
        for j, mask_img_ in enumerate(mask_images):
            if i == j:
                sig_sim.append(0)
                continue
            score, _ = compare_ssim(mask_img, mask_img_, win_size=3, full=True)
            sig_sim.append(score)
        similarity.append(np.array(sig_sim))

    similarity = np.array(similarity)
    sum_similarity = np.sum(similarity, axis=0)
    unique_index = np.argmin(sum_similarity)
    unique_circle = circles[unique_index]

    return unique_circle[0], unique_circle[1], unique_circle[2]


def find_unique_color(img: np.ndarray, circles: List[List[int]]) -> Tuple[int, int, int]:
    mask_images = _build_mask(img, circles, lookup="color")

    original_mask_images = mask_images.copy()
    original_mask_images = np.array(original_mask_images)

    mask_images = [cv2.convertScaleAbs(mask_img, alpha=1.5, beta=0) for mask_img in mask_images]

    scores = []

    for original_mask_img, mask_img in zip(original_mask_images, mask_images):
        gray = cv2.cvtColor(mask_img, cv2.COLOR_BGR2GRAY)
        center_color = gray[gray.shape[0] // 2, gray.shape[1] // 2]
        gray[gray < min(200, int(center_color))] = 0

        # remove noise
        kernel = np.ones((3, 3), np.uint8)
        gray = cv2.morphologyEx(gray, cv2.MORPH_OPEN, kernel)

        # binary
        threshold = 255
        gray[gray < threshold] = 0
        gray[gray >= threshold] = 255

        colors = original_mask_img[gray == 0]
        colors = np.array(colors).reshape(-1, 3)
        score = np.var(colors, axis=0).sum()

        scores.append(score)

    unique_index = np.argmin(scores)

    unique_circle = circles[unique_index]

    return unique_circle[0], unique_circle[1], unique_circle[2]



# -*- coding: utf-8 -*-
# Time       : 2023/10/20 17:28
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description: zero-shot image classification
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import List, Literal, Iterable, Tuple

import onnxruntime
from PIL.Image import Image

from hcaptcha_challenger.components.prompt_handler import handle
from hcaptcha_challenger.onnx.clip import MossCLIP
from hcaptcha_challenger.onnx.modelhub import ModelHub, DataLake
from hcaptcha_challenger.onnx.utils import is_cuda_pipline_available


def register_pipline(
    modelhub: ModelHub,
    *,
    fmt: Literal["onnx", "transformers"] = None,
    install_only: bool = False,
    **kwargs,
):
    """
    Ace Model:
        - laion/CLIP-ViT-L-14-DataComp.XL-s13B-b90K --> ONNX 1.7GB
        - QuanSun/EVA-CLIP/EVA02_CLIP_L_psz14_224to336 --> ONNX

    :param install_only:
    :param modelhub:
    :param fmt:
    :param kwargs:
    :return:
    """
    if fmt in ["transformers", None]:
        fmt = "transformers" if is_cuda_pipline_available else "onnx"

    if fmt in ["onnx"]:
        v_net, t_net = None, None

        if not modelhub.label_alias:
            modelhub.parse_objects()

        if visual_path := kwargs.get("visual_path"):
            if not isinstance(visual_path, Path):
                raise ValueError("visual_path should be a pathlib.Path")
            if not visual_path.is_file():
                raise FileNotFoundError(
                    "Select to use visual ONNX model, but the specified model does not exist -"
                    f" {visual_path=}"
                )
            v_net = onnxruntime.InferenceSession(
                visual_path, providers=onnxruntime.get_available_providers()
            )
        if textual_path := kwargs.get("textual_path"):
            if not isinstance(textual_path, Path):
                raise ValueError("textual_path should be a pathlib.Path")
            if not textual_path.is_file():
                raise FileNotFoundError(
                    "Select to use textual ONNX model, but the specified model does not exist -"
                    f" {textual_path=}"
                )
            t_net = onnxruntime.InferenceSession(
                textual_path, providers=onnxruntime.get_available_providers()
            )

        if not v_net:
            visual_model = kwargs.get("visual_model", modelhub.DEFAULT_CLIP_VISUAL_MODEL)
            v_net = modelhub.match_net(visual_model, install_only=install_only)
        if not t_net:
            textual_model = kwargs.get("textual_model", modelhub.DEFAULT_CLIP_TEXTUAL_MODEL)
            t_net = modelhub.match_net(textual_model, install_only=install_only)

        if not install_only:
            _pipeline = MossCLIP.from_pluggable_model(v_net, t_net)
            return _pipeline

    if fmt in ["transformers"]:
        from transformers import pipeline  # type:ignore

        import torch  # type:ignore

        device = "cuda" if torch.cuda.is_available() else "cpu"
        checkpoint = kwargs.get("checkpoint", "laion/CLIP-ViT-L-14-DataComp.XL-s13B-b90K")
        task = kwargs.get("task", "zero-shot-image-classification")
        batch_size = kwargs.get("batch_size", 16)
        _pipeline = pipeline(task=task, device=device, model=checkpoint, batch_size=batch_size)
        return _pipeline


def format_datalake(dl: DataLake) -> Tuple[List[str], List[str]]:
    positive_labels = dl.positive_labels.copy()
    negative_labels = dl.negative_labels.copy()

    # When the input is a challenge prompt, cut it into phrases
    if dl.raw_prompt:
        prompt = dl.raw_prompt
        prompt = prompt.replace("_", " ")
        true_label = handle(prompt)
        if true_label not in positive_labels:
            positive_labels.append(true_label)
        if not negative_labels:
            false_label = dl.PREMISED_BAD.format(true_label)
            negative_labels.append(false_label)

    # Insert hypothesis_template
    for labels in [positive_labels, negative_labels]:
        for i, label in enumerate(labels):
            if "This is a" in label:
                continue
            labels[i] = dl.PREMISED_YES.format(label)

    # Formatting model input
    candidate_labels = positive_labels.copy()
    if isinstance(negative_labels, list) and len(negative_labels) != 0:
        candidate_labels.extend(negative_labels)

    return positive_labels, candidate_labels


@dataclass
class ZeroShotImageClassifier:
    positive_labels: List[str] = field(default_factory=list)
    candidate_labels: List[str] = field(default_factory=list)

    @classmethod
    def from_datalake(cls, dl: DataLake):
        positive_labels, candidate_labels = format_datalake(dl)
        return cls(positive_labels=positive_labels, candidate_labels=candidate_labels)

    def __call__(self, detector: MossCLIP, image: Image, *args, **kwargs):
        if isinstance(detector, MossCLIP) and not isinstance(image, Iterable):
            image = [image]
        predictions = detector(image, candidate_labels=self.candidate_labels)
        return predictions



# -*- coding: utf-8 -*-
# Time       : 2023/10/15 21:37
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

import os

import cv2
import importlib_metadata
import numpy as np

USE_TF = os.environ.get("USE_TF", "AUTO").upper()
USE_TORCH = os.environ.get("USE_TORCH", "AUTO").upper()

ENV_VARS_TRUE_VALUES = {"1", "ON", "YES", "TRUE"}
ENV_VARS_TRUE_AND_AUTO_VALUES = ENV_VARS_TRUE_VALUES.union({"AUTO"})


def _is_package_available(pkg_name: str) -> bool:
    try:
        package_version = importlib_metadata.metadata(pkg_name) is not None
        return bool(package_version)
    except importlib_metadata.PackageNotFoundError:
        return False


_torch_available = False
if USE_TORCH in ENV_VARS_TRUE_AND_AUTO_VALUES and USE_TF not in ENV_VARS_TRUE_VALUES:
    _torch_available = _is_package_available("torch")
else:
    _torch_available = False

_transformers_available = _is_package_available("transformers")


def is_torch_available():
    return _torch_available


def is_transformers_available():
    return _transformers_available


is_cuda_pipline_available = is_torch_available() and is_transformers_available()


def nms(boxes, scores, iou_threshold):
    # Sort by score
    sorted_indices = np.argsort(scores)[::-1]

    keep_boxes = []
    while sorted_indices.size > 0:
        # Pick the last box
        box_id = sorted_indices[0]
        keep_boxes.append(box_id)

        # Compute IoU of the picked box with the rest
        ious = compute_iou(boxes[box_id, :], boxes[sorted_indices[1:], :])

        # Remove boxes with IoU over the threshold
        keep_indices = np.where(ious < iou_threshold)[0]

        # print(keep_indices.shape, sorted_indices.shape)
        sorted_indices = sorted_indices[keep_indices + 1]

    return keep_boxes


def multiclass_nms(boxes, scores, class_ids, iou_threshold):
    unique_class_ids = np.unique(class_ids)

    keep_boxes = []
    for class_id in unique_class_ids:
        class_indices = np.where(class_ids == class_id)[0]
        class_boxes = boxes[class_indices, :]
        class_scores = scores[class_indices]

        class_keep_boxes = nms(class_boxes, class_scores, iou_threshold)
        keep_boxes.extend(class_indices[class_keep_boxes])

    return keep_boxes


def compute_iou(box, boxes):
    # Compute xmin, ymin, xmax, ymax for both boxes
    xmin = np.maximum(box[0], boxes[:, 0])
    ymin = np.maximum(box[1], boxes[:, 1])
    xmax = np.minimum(box[2], boxes[:, 2])
    ymax = np.minimum(box[3], boxes[:, 3])

    # Compute intersection area
    intersection_area = np.maximum(0, xmax - xmin) * np.maximum(0, ymax - ymin)

    # Compute union area
    box_area = (box[2] - box[0]) * (box[3] - box[1])
    boxes_area = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])
    union_area = box_area + boxes_area - intersection_area

    # Compute IoU
    iou = intersection_area / union_area

    return iou


def xywh2xyxy(x):
    # Convert bounding box (x, y, w, h) to bounding box (x1, y1, x2, y2)
    y = np.copy(x)
    y[..., 0] = x[..., 0] - x[..., 2] / 2
    y[..., 1] = x[..., 1] - x[..., 3] / 2
    y[..., 2] = x[..., 0] + x[..., 2] / 2
    y[..., 3] = x[..., 1] + x[..., 3] / 2
    return y


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def draw_detections(
    image, boxes, scores, class_ids, colors, classes, mask_alpha=0.3, mask_maps=None
):
    img_height, img_width = image.shape[:2]
    size = min([img_height, img_width]) * 0.0006
    text_thickness = int(min([img_height, img_width]) * 0.001)

    mask_img = draw_masks(image, boxes, class_ids, colors, mask_alpha, mask_maps)

    # Draw bounding boxes and labels of detections
    for box, score, class_id in zip(boxes, scores, class_ids):
        color = colors[class_id]

        x1, y1, x2, y2 = box.astype(int)

        # Draw rectangle
        cv2.rectangle(mask_img, (x1, y1), (x2, y2), color, 2)

        label = classes[class_id]
        caption = f"{label} {int(score * 100)}%"
        (tw, th), _ = cv2.getTextSize(
            text=caption,
            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
            fontScale=size,
            thickness=text_thickness,
        )
        th = int(th * 1.2)

        cv2.rectangle(mask_img, (x1, y1), (x1 + tw, y1 - th), color, -1)

        cv2.putText(
            mask_img,
            caption,
            (x1, y1),
            cv2.FONT_HERSHEY_SIMPLEX,
            size,
            (255, 255, 255),
            text_thickness,
            cv2.LINE_AA,
        )

    return mask_img


def draw_masks(image, boxes, class_ids, colors, mask_alpha=0.3, mask_maps=None):
    mask_img = image.copy()

    # Draw bounding boxes and labels of detections
    for i, (box, class_id) in enumerate(zip(boxes, class_ids)):
        color = colors[class_id]

        x1, y1, x2, y2 = box.astype(int)

        # Draw fill mask image
        if mask_maps is None:
            cv2.rectangle(mask_img, (x1, y1), (x2, y2), color, -1)
        else:
            crop_mask = mask_maps[i][y1:y2, x1:x2, np.newaxis]
            crop_mask_img = mask_img[y1:y2, x1:x2]
            crop_mask_img = crop_mask_img * (1 - crop_mask) + crop_mask * color
            mask_img[y1:y2, x1:x2] = crop_mask_img

    return cv2.addWeighted(mask_img, mask_alpha, image, 1 - mask_alpha, 0)


def draw_comparison(img1, img2, name1, name2, fontsize=2.6, text_thickness=3):
    (tw, th), _ = cv2.getTextSize(
        text=name1, fontFace=cv2.FONT_HERSHEY_DUPLEX, fontScale=fontsize, thickness=text_thickness
    )
    x1 = img1.shape[1] // 3
    y1 = th
    offset = th // 5
    cv2.rectangle(
        img1,
        (x1 - offset * 2, y1 + offset),
        (x1 + tw + offset * 2, y1 - th - offset),
        (0, 115, 255),
        -1,
    )
    cv2.putText(
        img1, name1, (x1, y1), cv2.FONT_HERSHEY_DUPLEX, fontsize, (255, 255, 255), text_thickness
    )

    (tw, th), _ = cv2.getTextSize(
        text=name2, fontFace=cv2.FONT_HERSHEY_DUPLEX, fontScale=fontsize, thickness=text_thickness
    )
    x1 = img2.shape[1] // 3
    y1 = th
    offset = th // 5
    cv2.rectangle(
        img2,
        (x1 - offset * 2, y1 + offset),
        (x1 + tw + offset * 2, y1 - th - offset),
        (94, 23, 235),
        -1,
    )

    cv2.putText(
        img2, name2, (x1, y1), cv2.FONT_HERSHEY_DUPLEX, fontsize, (255, 255, 255), text_thickness
    )

    combined_img = cv2.hconcat([img1, img2])
    if combined_img.shape[1] > 3840:
        combined_img = cv2.resize(combined_img, (3840, 2160))

    return combined_img



# -*- coding: utf-8 -*-
# Time       : 2023/8/19 18:23
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

import gc
import json
import os
import shutil
import time
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from json import JSONDecodeError
from pathlib import Path
from typing import Dict, List, Tuple
from urllib.parse import urlparse

import cv2
import httpx
import onnxruntime
import yaml
from cv2.dnn import Net
from loguru import logger
from onnxruntime import InferenceSession
from tqdm import tqdm

from hcaptcha_challenger.utils import from_dict_to_model

this_dir: Path = Path(__file__).parent
DEFAULT_KEYPOINT_MODEL = "COCO2020_yolov8m.onnx"


@logger.catch
def request_resource(url: str, save_path: Path):
    cdn_prefix = os.environ.get("MODELHUB_CDN_PREFIX", "")

    if cdn_prefix and cdn_prefix.startswith("https://"):
        parser = urlparse(cdn_prefix)
        scheme, netloc = parser.scheme, parser.netloc
        url = f"{scheme}://{netloc}/{url}"

    headers = {
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/115.0.0.0 Safari/537.36 Edg/115.0.1901.203"
    }
    with open(save_path, "wb") as download_file:
        with httpx.Client(headers=headers, follow_redirects=True) as client:
            with client.stream("GET", url) as response:
                total = int(response.headers["Content-Length"])
                with tqdm(
                    total=total,
                    unit_scale=True,
                    unit_divisor=1024,
                    unit="B",
                    desc=f"Installing {save_path.parent.name}/{save_path.name}",
                ) as progress:
                    num_bytes_downloaded = response.num_bytes_downloaded
                    for chunk in response.iter_bytes():
                        download_file.write(chunk)
                        progress.update(response.num_bytes_downloaded - num_bytes_downloaded)
                        num_bytes_downloaded = response.num_bytes_downloaded


@dataclass
class ReleaseAsset:
    id: int
    node_id: str
    name: str
    size: int
    browser_download_url: str


@dataclass
class Assets:
    """
    ONNX model manager
    """

    release_url: str
    """
    GitHub Release URL
    Such as https://api.github.com/repos/QIN2DIM/hcaptcha-challenger/releases
    """

    _assets_dir: Path = this_dir.joinpath("models/_assets")
    """
    The local path of index file
    """

    _memory_dir: Path = this_dir.joinpath("models/_memory")
    """
    Archive historical versions of a model
    """

    cache_lifetime: timedelta = timedelta(hours=2)
    """
    The effective time of the index file
    """

    _name2asset: Dict[str, ReleaseAsset] = field(default_factory=dict)
    """
    { model_name.onnx: {ReleaseAsset} }
    """

    _name2node: Dict[str, str] = field(default_factory=dict)
    """
    model_name.onnx to asset_node_id
    """

    def __post_init__(self):
        for ck in [self._assets_dir, self._memory_dir]:
            ck.mkdir(mode=0o777, parents=True, exist_ok=True)

    @classmethod
    def from_release_url(cls, release_url: str, **kwargs):
        instance = cls(release_url=release_url, **kwargs)

        # Assets - latest information
        # Called only before the Assets._pull network request is initiated,
        # and the effective local cache will replace the remote resource
        assets = [instance._assets_dir.joinpath(i) for i in os.listdir(instance._assets_dir)]
        if assets:
            resp_path = assets[-1]
            resp_ctime = datetime.fromtimestamp(resp_path.stat().st_ctime)
            if resp_ctime + instance.cache_lifetime > datetime.now():
                try:
                    data = json.loads(resp_path.read_text(encoding="utf8"))
                    instance._name2asset = {
                        k: from_dict_to_model(ReleaseAsset, v) for k, v in data.items()
                    }
                except JSONDecodeError as err:
                    logger.warning(err)

        # Memory - version control
        for x in os.listdir(instance._memory_dir):
            name, node = x.rsplit(".", maxsplit=1)
            instance._name2node[name] = node

        return instance

    def get_focus_asset(self, focus_name: str) -> ReleaseAsset:
        return self._name2asset.get(focus_name)

    def flush_runtime_assets(self, upgrade: bool = False):
        """Request assets index from remote repository"""
        self._name2asset = {}

        if upgrade is True:
            shutil.rmtree(self._assets_dir, ignore_errors=True)
            self._assets_dir.mkdir(parents=True, exist_ok=True)

        assets_paths = [self._assets_dir.joinpath(i) for i in os.listdir(self._assets_dir)]
        is_outdated = False
        if assets_paths:
            resp_path = assets_paths[-1]
            resp_ctime = datetime.fromtimestamp(resp_path.stat().st_ctime)
            if resp_ctime + self.cache_lifetime < datetime.now():
                is_outdated = True

        # Request assets index from remote repository
        if upgrade is True or not assets_paths or is_outdated:
            try:
                resp = httpx.get(self.release_url, timeout=3)
                data = resp.json()[0]
                assets: List[dict] = data.get("assets", [])
                for asset in assets:
                    release_asset = from_dict_to_model(ReleaseAsset, asset)
                    self._name2asset[asset["name"]] = release_asset
            except (httpx.ConnectError, JSONDecodeError) as err:
                logger.error(err)
            except (AttributeError, IndexError, KeyError) as err:
                logger.error(err)

        # Only implemented after the Assets._pull network request is initiated,
        # it is used to update the content and timestamp of the local cache
        for asset_fn in os.listdir(self._assets_dir):
            asset_src = self._assets_dir.joinpath(asset_fn)
            asset_dst = self._assets_dir.joinpath(f"_{asset_fn.replace('_', '')}")
            shutil.move(asset_src, asset_dst)
        assets_path = self._assets_dir.joinpath(f"{int(time.time())}.json")
        data = {k: v.__dict__ for k, v in self._name2asset.items()}
        assets_path.write_text(json.dumps(data, ensure_ascii=True, indent=2))

    def archive_memory(self, focus_name: str, new_node_id: str):
        """
        Archive model version to local
        :param focus_name: model name with .onnx suffix
        :param new_node_id: node_id of the GitHub release, ONNX model file.
        :return:
        """
        old_node_id = self._name2node.get(focus_name, "")

        self._name2node[focus_name] = new_node_id

        # Create or update a history tracking node
        if not old_node_id:
            memory_node = self._memory_dir.joinpath(f"{focus_name}.{new_node_id}")
            memory_node.write_text(str(memory_node))
        else:
            memory_src = self._memory_dir.joinpath(f"{focus_name}.{old_node_id}")
            memory_dst = self._memory_dir.joinpath(f"{focus_name}.{new_node_id}")
            shutil.move(memory_src, memory_dst)

    def is_outdated(self, focus_name: str):
        """
        Diagnostic steps for delayed reflection to maintain consistency at both ends of a distributed network
        :param focus_name: model name with .onnx suffix
        :return:
        """
        focus_asset = self._name2asset.get(focus_name)
        if not focus_asset:
            return

        local_node_id = self._name2node.get(focus_name, "")
        if not local_node_id:
            return

        if local_node_id != focus_asset.node_id:
            return True
        return False


@dataclass
class ModelHub:
    """
    Manage pluggable models. Provides high-level interfaces
    such as model download, model cache, and model scheduling.
    """

    models_dir = this_dir.joinpath("models")
    assets_dir = models_dir.joinpath("_assets")
    objects_path = models_dir.joinpath("objects.yaml")

    lang: str = "en"

    label_alias: Dict[str, str] = field(default_factory=dict)
    """
    Image classification
    ---
    The most basic function
    Storing a series of mappings from model names to short prompts, 
    I .e., what model to use to handle what challenge is determined by this dictionary.
    """

    yolo_names: List[str] = field(default_factory=list)
    ashes_of_war: Dict[str, List[str]] = field(default_factory=dict)
    """
    Object Detection
    ---
    Provide a series of object detection models applied to special tasks.
    The yolo_names stores the label names of all task objects that the model can process.
    """

    nested_categories: Dict[str, List[str]] = field(default_factory=dict)
    """
    Model Rank.Strategy
    ---
    Provide a string of small model clusters for a prompt to realize 
    "find the ｛z｝ pictures most similar to ｛y｝ in the ｛x_i｝ pictures"
    """

    circle_segment_model: str = field(default=str)
    """
    Image Segmentation
    ---
    A model trained specifically for image segmentation tasks 
    that can separate background and foreground with close to 100 percent accuracy
    """

    datalake: Dict[str, DataLake] = field(default_factory=dict)
    """
    ViT zero-shot image classification
    ---
    Used to generate prompt templates to intensify inserted CLIP model and improve accuracy.
    """

    DEFAULT_CLIP_VISUAL_MODEL: str = "visual_CLIP_RN50.openai.onnx"
    DEFAULT_CLIP_TEXTUAL_MODEL: str = "textual_CLIP_RN50.openai.onnx"
    """
    Available Model
    --- 1180+ MiB
    DEFAULT_CLIP_VISUAL_MODEL: str = "visual_CLIP_ViT-B-32.openai.onnx"
    DEFAULT_CLIP_TEXTUAL_MODEL: str = "textual_CLIP_ViT-B-32.openai.onnx"
    --- 658.3 MiB
    DEFAULT_CLIP_VISUAL_MODEL: str = "visual_CLIP_RN50.openai.onnx"
    DEFAULT_CLIP_TEXTUAL_MODEL: str = "textual_CLIP_RN50.openai.onnx"
    --- 3300+ MiB
    DEFAULT_CLIP_VISUAL_MODEL: str = "visual_CLIP-ViT-L-14-DataComp.XL-s13B-b90K.onnx"
    DEFAULT_CLIP_TEXTUAL_MODEL: str = "textual_CLIP-ViT-L-14-DataComp.XL-s13B-b90K.onnx"
    """

    release_url: str = ""
    objects_url: str = ""

    assets: Assets = None

    _name2net: Dict[str, Net | InferenceSession] = field(default_factory=dict)
    """
    { model_name1.onnx: cv2.dnn.Net }
    { model_name2.onnx: onnxruntime.InferenceSession }
    """

    def __post_init__(self):
        self.assets_dir.mkdir(mode=0o777, parents=True, exist_ok=True)

    @classmethod
    def from_github_repo(cls, username: str = "QIN2DIM", lang: str = "en", **kwargs):
        release_url = f"https://api.github.com/repos/{username}/hcaptcha-challenger/releases"
        objects_url = f"https://raw.githubusercontent.com/{username}/hcaptcha-challenger/main/src/objects.yaml"

        instance = cls(release_url=release_url, objects_url=objects_url, lang=lang)
        instance.assets = Assets.from_release_url(release_url)

        return instance

    def pull_objects(self, upgrade: bool = False):
        """Network request"""
        if (
            upgrade
            or not self.objects_path.exists()
            or not self.objects_path.stat().st_size
            or time.time() - self.objects_path.stat().st_mtime > 3600
        ):
            request_resource(self.objects_url, self.objects_path)

    def parse_objects(self):
        """Try to load label_alias from local database"""
        if not self.objects_path.exists():
            return

        data = yaml.safe_load(self.objects_path.read_text(encoding="utf8"))
        if not data:
            os.remove(self.objects_path)
            return

        label_to_i18n_mapping: dict = data.get("label_alias", {})
        if label_to_i18n_mapping:
            for model_name, lang_to_prompts in label_to_i18n_mapping.items():
                for lang, prompts in lang_to_prompts.items():
                    if lang != self.lang:
                        continue
                    self.label_alias.update({prompt.strip(): model_name for prompt in prompts})

        yolo2names: Dict[str, List[str]] = data.get("ashes_of_war", {})
        if yolo2names:
            self.yolo_names = [cl for cc in yolo2names.values() for cl in cc]
            self.ashes_of_war = yolo2names

        nested_categories = data.get("nested_categories", {})
        if nested_categories:
            self.nested_categories = nested_categories

        self.circle_segment_model = data.get(
            "circle_seg", "appears_only_once_2309_yolov8s-seg.onnx"
        )

        datalake = data.get("datalake", {})
        if datalake:
            for prompt, serialized_binary in datalake.items():
                datalake[prompt] = DataLake.from_serialized(serialized_binary)
        self.datalake = datalake

    def pull_model(self, focus_name: str):
        """
        1. node_id: Record the insertion point
        and indirectly judge the changes of the file with the same name

        2. assets.List: Record the item list of the release attachment,
        and directly determine whether there are undownloaded files

        3. assets.size: Record the amount of bytes inserted into the file,
        and directly determine whether the file is downloaded completely

        :param focus_name: model_name.onnx  Such as `mobile.onnx`
        :return:
        """
        focus_asset = self.assets.get_focus_asset(focus_name)
        if not focus_asset:
            return

        # Matching conditions to trigger download tasks
        model_path = self.models_dir.joinpath(focus_name)
        if (
            not model_path.exists()
            or model_path.stat().st_size != focus_asset.size
            or self.assets.is_outdated(focus_name)
        ):
            request_resource(focus_asset.browser_download_url, model_path.absolute())
            self.assets.archive_memory(focus_name, focus_asset.node_id)

    def active_net(self, focus_name: str) -> Net | InferenceSession | None:
        """Load and register an existing model"""
        model_path = self.models_dir.joinpath(focus_name)
        if (
            model_path.exists()
            and model_path.stat().st_size
            and not self.assets.is_outdated(focus_name)
        ):
            if "yolo" in focus_name.lower() or "clip" in focus_name.lower():
                net = onnxruntime.InferenceSession(
                    model_path, providers=onnxruntime.get_available_providers()
                )
            else:
                net = cv2.dnn.readNetFromONNX(str(model_path))
            self._name2net[focus_name] = net
            return net

    def match_net(
        self, focus_name: str, *, install_only: bool = False
    ) -> Net | InferenceSession | None:
        """
        When a PluggableONNXModel object is instantiated:
        ---

        - It automatically reads and registers model objects specified in objects.yaml
        that already exist in the designated directory.
        - However, the model files corresponding to the label groups expressed in objects.yaml
        do not necessarily all exist yet.
        - No new network requests are made during initialization,
        i.e. missing models are not downloaded during the initialization phase.

        match_net models are passively pulled:
        ---

        - Missing ONNX models used for handling specific binary classification tasks are
        passively downloaded during the challenge.
        - Matching models are automatically downloaded, registered, and returned.
        - Models not on the objects.yaml list will not be downloaded.

        [!] The newly inserted model can be used directly.
        :param install_only:
        :param focus_name: model_name with .onnx suffix
        :return:
        """
        net = self._name2net.get(focus_name)
        if not net:
            self.pull_model(focus_name)
            if not install_only:
                net = self.active_net(focus_name)
        return net

    def unplug(self):
        for ash in self.ashes_of_war:
            if ash not in self._name2net:
                continue
            del self._name2net[ash]
            gc.collect()

        for m in [self.DEFAULT_CLIP_TEXTUAL_MODEL, self.DEFAULT_CLIP_VISUAL_MODEL]:
            if m in self._name2net:
                del self._name2net[m]
                gc.collect()

    def apply_ash_of_war(self, ash: str) -> Tuple[str, List[str]]:
        # Prelude - pending DensePose
        if "head of " in ash and "animal" in ash:
            for model_name, covered_class in self.ashes_of_war.items():
                if "head" not in model_name:
                    continue
                for class_name in covered_class:
                    if class_name.replace("-head", "") in ash:
                        return model_name, covered_class

        # Prelude - Ordered dictionary
        for model_name, covered_class in self.ashes_of_war.items():
            for class_name in covered_class:
                if class_name in ash:
                    return model_name, covered_class

        # catch-all rules
        return DEFAULT_KEYPOINT_MODEL, self.ashes_of_war[DEFAULT_KEYPOINT_MODEL]

    def lookup_ash_of_war(self, ash: str):  # fixme
        """catch-all default cases"""
        if "can be eaten" in ash:
            for model_name, covered_class in self.ashes_of_war.items():
                if "can_be_eaten" in model_name:
                    yield model_name, covered_class

        if "not an animal" in ash:
            for model_name, covered_class in self.ashes_of_war.items():
                if "notanimal" in model_name:
                    yield model_name, covered_class

        if "head of " in ash and "animal" in ash:
            for model_name, covered_class in self.ashes_of_war.items():
                if "head" in model_name:
                    yield model_name, covered_class

        if "animal" in ash and "not belong to the sea" in ash:
            for model_name, covered_class in self.ashes_of_war.items():
                if (
                    "notseaanimal" in model_name
                    or "fantasia_elephant" in model_name
                    or "fantasia_cat" in model_name
                ):
                    yield model_name, covered_class

        for model_name, covered_class in self.ashes_of_war.items():
            binder = model_name.split("_")
            if len(binder) > 2 and binder[-2].isdigit():
                binder = " ".join(model_name.split("_")[:-2])
                if binder in ash:
                    yield model_name, covered_class
            else:
                for class_name in covered_class:
                    if class_name in ash:
                        yield model_name, covered_class


@dataclass
class DataLake:
    positive_labels: List[str] = field(default_factory=list)
    """
    Indicate the label with the meaning "True", 
    preferably an independent noun or clause
    """

    negative_labels: List[str] = field(default_factory=list)
    """
    Indicate the label with the meaning "False", 
    preferably an independent noun or clause
    """

    joined_dirs: List[str] | Path | None = None
    """
    Attributes reserved for AutoLabeling
    Used to indicate the directory where the dataset is located

    input_dir = db_dir.joinpath(*joined_dirs).absolute()
    """

    raw_prompt: str = ""
    """
    Challenge prompt or keywords after being divided
    
    !! IMPORT !!
    Only for unsupervised challenges.
    Please do not read in during the initialization phase.
    """

    PREMISED_YES: str = "This is a picture that looks like {}."
    PREMISED_BAD: str = "This is a picture that don't look like {}."
    """
    Insert self-supervised prompt
    """

    @classmethod
    def from_challenge_prompt(cls, raw_prompt: str):
        return cls(raw_prompt=raw_prompt)

    @classmethod
    def from_serialized(cls, fields: Dict[str, List[str]]):
        positive_labels = []
        negative_labels = []
        for kb, labels in fields.items():
            kb = kb.lower()
            if "pos" in kb or kb.startswith("t"):
                positive_labels = labels
            elif "neg" in kb or kb.startswith("f"):
                negative_labels = labels
        return cls(positive_labels=positive_labels, negative_labels=negative_labels)

    @classmethod
    def from_binary_labels(cls, positive_labels: List[str], negative_labels: List[str]):
        return cls(positive_labels=positive_labels, negative_labels=negative_labels)



# -*- coding: utf-8 -*-
# Time       : 2022/3/2 0:52
# Author     : QIN2DIM
# Github     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

import math
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import List
from typing import Literal

import cv2
import numpy as np
from onnxruntime import InferenceSession

from .utils import nms, xywh2xyxy, draw_detections, sigmoid, multiclass_nms


@dataclass
class YOLOv8:
    conf_threshold: float = 0.5
    iou_threshold: float = 0.5
    classes: List[str] = field(default_factory=list)

    session: InferenceSession = None

    input_names = None
    input_shape = None
    input_height = None
    input_width = None
    output_names = None

    img_height = None
    img_width = None

    def __post_init__(self):
        model_inputs = self.session.get_inputs()
        self.input_names = [model_inputs[i].name for i in range(len(model_inputs))]

        self.input_shape = model_inputs[0].shape
        self.input_height = self.input_shape[2]
        self.input_width = self.input_shape[3]

        model_outputs = self.session.get_outputs()
        self.output_names = [model_outputs[i].name for i in range(len(model_outputs))]

    @classmethod
    def from_pluggable_model(cls, session: InferenceSession, classes: List[str]):
        return cls(session=session, classes=classes)

    def __call__(self, image: Path | bytes, shape_type: Literal["point", "bounding_box"] = "point"):
        if isinstance(image, Path):
            image = image.read_bytes()

        np_array = np.frombuffer(image, np.uint8)
        image = cv2.imdecode(np_array, flags=1)

        boxes, scores, class_ids = self.detect_objects(image)

        response = []
        if shape_type == "point":
            for i, class_id in enumerate(class_ids):
                x1, y1, x2, y2 = boxes[i]
                center_x, center_y = int((x2 - x1) / 2 + x1), int((y2 - y1) / 2 + y1)
                response.append((self.classes[class_id], (center_x, center_y), scores[i]))
        elif shape_type == "bounding_box":
            for i, class_id in enumerate(class_ids):
                x1, y1, x2, y2 = boxes[i]
                point_start, point_end = (x1, y1), (x2, y2)
                response.append((self.classes[class_id], point_start, point_end, scores[i]))
        return response

    def detect_objects(self, image: np.ndarray):
        self.img_height, self.img_width = image.shape[:2]

        input_tensor = self._prepare_input(image)

        # Perform inference on the image
        outputs = self.session.run(self.output_names, {self.input_names[0]: input_tensor})

        boxes, scores, class_ids = self._process_output(outputs)

        return boxes, scores, class_ids

    def _prepare_input(self, image):
        input_img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Resize input image
        input_img = cv2.resize(input_img, (self.input_width, self.input_height))

        # Scale input pixel values to 0 to 1
        input_img = input_img / 255.0
        input_img = input_img.transpose(2, 0, 1)
        input_tensor = input_img[np.newaxis, :, :, :].astype(np.float32)

        return input_tensor

    def _process_output(self, output):
        predictions = np.squeeze(output[0]).T

        # Filter out object confidence scores below threshold
        scores = np.max(predictions[:, 4:], axis=1)
        predictions = predictions[scores > self.conf_threshold, :]
        scores = scores[scores > self.conf_threshold]

        if len(scores) == 0:
            return [], [], []

        # Get the class with the highest confidence
        class_ids = np.argmax(predictions[:, 4:], axis=1)

        # Get bounding boxes for each object
        # Extract boxes from predictions
        boxes = predictions[:, :4]

        # Scale boxes to original image dimensions
        # Rescale boxes to original image dimensions
        input_shape = np.array(
            [self.input_width, self.input_height, self.input_width, self.input_height]
        )
        boxes = np.divide(boxes, input_shape, dtype=np.float32)
        boxes *= np.array([self.img_width, self.img_height, self.img_width, self.img_height])

        # Convert boxes to xyxy format
        boxes = xywh2xyxy(boxes)

        # Apply non-maxima suppression to suppress weak, overlapping bounding boxes
        # indices = nms(boxes, scores, self.iou_threshold)
        indices = multiclass_nms(boxes, scores, class_ids, self.iou_threshold)

        return boxes[indices], scores[indices], class_ids[indices]


def is_matched_ash_of_war(ash: str, class_name: str):
    if "head of " in ash:
        if "head" not in class_name:
            return False
        keyword = class_name.replace("-head", "").strip()
        if keyword not in ash:
            return False
        return True

    # catch-all rules
    if class_name not in ash:
        return False
    return True


def finetune_keypoint(name: str, point: List[int, int]) -> List[int, int]:
    point = point.copy()
    if name in ["nine", "9"]:
        point[-1] = point[-1] + 8
        point[0] = point[0] + 2
    if name in ["two", "2"]:
        point[-1] = point[-1] + 7
        point[0] = point[0] + 4
    return point


@dataclass
class YOLOv8Seg:
    conf_threshold: float = 0.71
    iou_threshold: float = 0.5
    num_masks: int = 32
    classes: List[str] = field(default_factory=list)
    session: InferenceSession = None

    input_names = None
    input_shape = None
    input_height = None
    input_width = None
    output_names = None

    img_height = None
    img_width = None
    boxes = None
    scores = None
    class_ids = None
    mask_maps = None

    colors = None

    def __post_init__(self):
        model_inputs = self.session.get_inputs()
        self.input_names = [model_inputs[i].name for i in range(len(model_inputs))]

        self.input_shape = model_inputs[0].shape
        self.input_height = self.input_shape[2]
        self.input_width = self.input_shape[3]

        model_outputs = self.session.get_outputs()
        self.output_names = [model_outputs[i].name for i in range(len(model_outputs))]

        if not self.colors:
            rng = np.random.default_rng(3)
            self.colors = rng.uniform(0, 255, size=(len(self.classes), 3))

    @classmethod
    def from_pluggable_model(cls, session: InferenceSession, classes: List[str]):
        return cls(session=session, classes=classes)

    def __call__(
        self, image: Path | bytes, shape_type: Literal["point", "bounding_box"] = "point", **kwargs
    ):
        if isinstance(image, Path):
            image = image.read_bytes()

        np_array = np.frombuffer(image, np.uint8)
        image = cv2.imdecode(np_array, flags=1)

        boxes, scores, class_ids, mask_maps = self.segment_objects(image)

        response = []
        if shape_type == "point":
            for i, class_id in enumerate(class_ids):
                x1, y1, x2, y2 = boxes[i]
                center_x, center_y = int((x2 - x1) / 2 + x1), int((y2 - y1) / 2 + y1)
                response.append((self.classes[class_id], (center_x, center_y), scores[i]))
        elif shape_type == "bounding_box":
            for i, class_id in enumerate(class_ids):
                x1, y1, x2, y2 = boxes[i]
                point_start, point_end = (x1, y1), (x2, y2)
                response.append((self.classes[class_id], point_start, point_end, scores[i]))
        return response

    def segment_objects(self, image):
        input_tensor = self.prepare_input(image)

        # Perform inference on the image
        outputs = self.inference(input_tensor)

        self.boxes, self.scores, self.class_ids, mask_pred = self.process_box_output(outputs[0])
        self.mask_maps = self.process_mask_output(mask_pred, outputs[1])

        return self.boxes, self.scores, self.class_ids, self.mask_maps

    def prepare_input(self, image):
        self.img_height, self.img_width = image.shape[:2]

        input_img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Resize input image
        input_img = cv2.resize(input_img, (self.input_width, self.input_height))

        # Scale input pixel values to 0 to 1
        input_img = input_img / 255.0
        input_img = input_img.transpose(2, 0, 1)
        input_tensor = input_img[np.newaxis, :, :, :].astype(np.float32)

        return input_tensor

    def inference(self, input_tensor):
        start = time.perf_counter()
        outputs = self.session.run(self.output_names, {self.input_names[0]: input_tensor})

        # print(f"Inference time: {(time.perf_counter() - start)*1000:.2f} ms")
        return outputs

    def process_box_output(self, box_output):
        predictions = np.squeeze(box_output).T
        num_classes = box_output.shape[1] - self.num_masks - 4

        # Filter out object confidence scores below threshold
        scores = np.max(predictions[:, 4 : 4 + num_classes], axis=1)
        predictions = predictions[scores > self.conf_threshold, :]
        scores = scores[scores > self.conf_threshold]

        if len(scores) == 0:
            return [], [], [], np.array([])

        box_predictions = predictions[..., : num_classes + 4]
        mask_predictions = predictions[..., num_classes + 4 :]

        # Get the class with the highest confidence
        class_ids = np.argmax(box_predictions[:, 4:], axis=1)

        # Get bounding boxes for each object
        boxes = self.extract_boxes(box_predictions)

        # Apply non-maxima suppression to suppress weak, overlapping bounding boxes
        indices = nms(boxes, scores, self.iou_threshold)

        return boxes[indices], scores[indices], class_ids[indices], mask_predictions[indices]

    def process_mask_output(self, mask_predictions, mask_output):
        if mask_predictions.shape[0] == 0:
            return []

        mask_output = np.squeeze(mask_output)

        # Calculate the mask maps for each box
        num_mask, mask_height, mask_width = mask_output.shape  # CHW
        masks = sigmoid(mask_predictions @ mask_output.reshape((num_mask, -1)))
        masks = masks.reshape((-1, mask_height, mask_width))

        # Downscale the boxes to match the mask size
        scale_boxes = self.rescale_boxes(
            self.boxes, (self.img_height, self.img_width), (mask_height, mask_width)
        )

        # For every box/mask pair, get the mask map
        mask_maps = np.zeros((len(scale_boxes), self.img_height, self.img_width))
        blur_size = (int(self.img_width / mask_width), int(self.img_height / mask_height))
        for i in range(len(scale_boxes)):
            scale_x1 = int(math.floor(scale_boxes[i][0]))
            scale_y1 = int(math.floor(scale_boxes[i][1]))
            scale_x2 = int(math.ceil(scale_boxes[i][2]))
            scale_y2 = int(math.ceil(scale_boxes[i][3]))

            x1 = int(math.floor(self.boxes[i][0]))
            y1 = int(math.floor(self.boxes[i][1]))
            x2 = int(math.ceil(self.boxes[i][2]))
            y2 = int(math.ceil(self.boxes[i][3]))

            scale_crop_mask = masks[i][scale_y1:scale_y2, scale_x1:scale_x2]
            crop_mask = cv2.resize(
                scale_crop_mask, (x2 - x1, y2 - y1), interpolation=cv2.INTER_CUBIC
            )

            crop_mask = cv2.blur(crop_mask, blur_size)

            crop_mask = (crop_mask > 0.5).astype(np.uint8)
            mask_maps[i, y1:y2, x1:x2] = crop_mask

        return mask_maps

    def extract_boxes(self, box_predictions):
        # Extract boxes from predictions
        boxes = box_predictions[:, :4]

        # Scale boxes to original image dimensions
        boxes = self.rescale_boxes(
            boxes, (self.input_height, self.input_width), (self.img_height, self.img_width)
        )

        # Convert boxes to xyxy format
        boxes = xywh2xyxy(boxes)

        # Check the boxes are within the image
        boxes[:, 0] = np.clip(boxes[:, 0], 0, self.img_width)
        boxes[:, 1] = np.clip(boxes[:, 1], 0, self.img_height)
        boxes[:, 2] = np.clip(boxes[:, 2], 0, self.img_width)
        boxes[:, 3] = np.clip(boxes[:, 3], 0, self.img_height)

        return boxes

    def draw_detections(self, image, mask_alpha=0.4):
        return draw_detections(
            image, self.boxes, self.scores, self.class_ids, self.colors, self.classes, mask_alpha
        )

    def draw_masks(self, image, mask_alpha=0.5):
        return draw_detections(
            image,
            self.boxes,
            self.scores,
            self.class_ids,
            self.colors,
            self.classes,
            mask_alpha,
            mask_maps=self.mask_maps,
        )

    @staticmethod
    def rescale_boxes(boxes, input_shape, image_shape):
        # Rescale boxes to original image dimensions
        input_shape = np.array([input_shape[1], input_shape[0], input_shape[1], input_shape[0]])
        boxes = np.divide(boxes, input_shape, dtype=np.float32)
        boxes *= np.array([image_shape[1], image_shape[0], image_shape[1], image_shape[0]])

        return boxes



# -*- coding: utf-8 -*-
# Time       : 2022/4/30 17:00
# Author     : Bingjie Yan
# Github     : https://github.com/beiyuouo
# Description:
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Tuple

import cv2
import numpy as np
from cv2.dnn import Net
from loguru import logger
from scipy.special import expit


class ChallengeStyle:
    WATERMARK = 144  # onTrigger 128x128
    GENERAL = 128
    GAN = 144


@dataclass
class ResNetControl:
    net: Net

    @classmethod
    def from_pluggable_model(cls, net: Net):
        return cls(net=net)

    def binary_classify(self, img_stream: Any) -> bool | Tuple[bool, np.ndarray]:
        img_arr = np.frombuffer(img_stream, np.uint8)
        img = cv2.imdecode(img_arr, flags=1)

        if img.shape[0] == ChallengeStyle.WATERMARK:
            img = cv2.fastNlMeansDenoisingColored(img, None, 10, 10, 7, 21)

        img = cv2.resize(img, (64, 64))
        blob = cv2.dnn.blobFromImage(img, 1 / 255.0, (64, 64), (0, 0, 0), swapRB=True, crop=False)

        # Use the delayed reflection mechanism
        # to ensure the consistency of both ends of the distributed network
        if self.net is None:
            logger.debug("The remote network does not exist or the local cache has expired.")
            return False
        self.net.setInput(blob)
        out = self.net.forward()
        logit = out[0]
        proba = expit(logit)
        if not np.argmax(out, axis=1)[0]:
            return True, proba
        return False, proba

    def execute(self, img_stream: bytes, **kwargs) -> bool | Tuple[bool, np.ndarray]:
        """Implementation process of solution"""
        get_proba = kwargs.get("proba", False)
        result = self.binary_classify(img_stream)
        if result is None or isinstance(result, bool):
            return result
        if isinstance(result, tuple):
            prediction, confidence = result
            if get_proba is True:
                return prediction, confidence
            return prediction



# -*- coding: utf-8 -*-
# Time       : 2023/10/9 16:18
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from pathlib import Path
from typing import List

import cv2
from sklearn.cluster import SpectralClustering


def get_2d_image(path: Path):
    image = cv2.imread(str(path))
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    return gray


def extract_features(img):
    hog = cv2.HOGDescriptor()
    h = hog.compute(img)
    return h


def find_similar_objects(example_paths: List[Path], challenge_paths: List[Path]):
    example_num = len(example_paths)

    results: List[bool | None] = [None] * len(challenge_paths)

    images_example = [get_2d_image(path) for path in example_paths]
    example_shape = images_example[0].shape[0]

    images_challenge = []
    for i, path in enumerate(challenge_paths):
        image = get_2d_image(path)
        if image.shape[0] != example_shape:
            results[i] = False
        images_challenge.append(image)

    images_merged = images_example + images_challenge

    X = [extract_features(img) for img in images_merged]

    clf = SpectralClustering(n_clusters=2, affinity="nearest_neighbors", n_neighbors=5)
    y = clf.fit_predict(X)

    ref_img_idx = 0
    ref_label = y[ref_img_idx]

    sim_idx_sequence = [i for i, rl in enumerate(y) if rl == ref_label]

    for idx in sim_idx_sequence[example_num:]:
        fit_idx = idx - example_num
        if results[fit_idx] is None:
            results[fit_idx] = True

    return results



# -*- coding: utf-8 -*-
# Time       : 2023/10/9 16:17
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from .appears_only_once import annotate_objects, find_unique_object, find_unique_color
from .largest_animal import find_similar_objects

__all__ = ["annotate_objects", "find_unique_object", "find_similar_objects", "find_unique_color"]
